"""
Copyright (c) 2009 [name + "(at)student.hpi.uni-potsdam.de" for name in ["murat.knecht", "franziska.haeger", "christian.schwarz"]]

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import sys, pygame, time, os
import sqlite3.dbapi2 as db

sys.path.append("..")
from gesture import Gesture
from vector2D import Vector2D

class GestureManager(object):
    
    def __init__(self, database=None, detaillevel=32):
        object.__init__(self)
        
        self.detaillevel = detaillevel
        self.gestures = []
        self.connection = self.get_database_connection(database)
        
        if self.connection:
            self.detaillevel = self.get_detaillevel_from_database()
            self.read_gestures_from_database()
        
        self.reset_or_create_new_gesture()
    
    def __del__(self):
        self.connection.commit()
        self.connection.close()
    
    def execute_sql(self, statement, returnResult=False):
        cursor = self.connection.cursor()
        
        if isinstance(statement, basestring):
            cursor.execute(statement)
        elif isinstance(statement, list):
            for sql in statement:
                cursor.execute(sql)
        
        self.connection.commit()
        
        result = None
        if returnResult:
            result = cursor.fetchall()
        
        cursor.close()
        return result
    
    def create_database_tables(self):
        sqls = []
        
        sql = "CREATE TABLE gestures (id INTEGER PRIMARY KEY, tolerance REAL"
        for i in xrange(self.detaillevel):
            sql += ", point_%s_x REAL, point_%s_y REAL" % (i, i)
        sql += ");"
        sqls.append(sql)
        
        sql = "CREATE TABLE metainfo (detaillevel INTEGER);"
        sqls.append(sql)
        
        sql = "INSERT INTO metainfo VALUES (%s);" % self.detaillevel
        sqls.append(sql)
        
        self.execute_sql(sqls)
    
    def get_database_connection(self, database=None):
        if not database:
            return None
        
        if not os.path.exists(database):
            self.connection = db.connect(database)
            self.create_database_tables()
            self.connection.close()            
        
        return db.connect(database)
    
    def get_detaillevel_from_database(self):
        sql = "SELECT detaillevel FROM metainfo;"
        return self.execute_sql(sql, True)[0][0]
    
    def read_gestures_from_database(self):
        sql = "SELECT * from gestures;"
        dbgestures = self.execute_sql(sql, True)
        
        for dbgesture in dbgestures:
            # dbgesture contains 2 * deteillevel + 2 elements (id and tolerance)
            assert(len(dbgesture) == 2 * self.detaillevel + 2)
            
            tolerance = dbgesture[1]
            
            gesture = Gesture()
            for idx in xrange(2, 2*self.detaillevel, 2):
                gesture.add_point((dbgesture[idx], dbgesture[idx+1]))
            
            gesture.finalize(self.detaillevel)
            self.gestures.append(gesture)
    
    def persist_gestures(self, event=None):
        if self.connection == None:
            return
        
        self.execute_sql("DELETE FROM gestures;")
        
        for id in xrange(len(self.gestures)):
            self.add_gesture_to_database(id, self.gestures[id])
        
        return True
    
    def add_gesture_to_database(self, id, gesture):
        if not gesture.finalized:
            gesture.finalize(self.detaillevel)
        
        assert(len(gesture) == self.detaillevel)
        
        sql = "INSERT INTO gestures VALUES (%s, %s" % (id, gesture.tolerance)
        for point in gesture.to_tuple_list():
            sql += ", %s, %s" % (point[0], point[1])
        sql+=");"
        
        self.execute_sql(sql)
    
    def add_gesture(self, gesture):
        if not gesture.finalized:
            gesture.finalize(self.detaillevel)
        
        self.tempGestures.append(gesture)
    
    def finalize(self):
        self.gestures.append(self.get_interpolated_gesture())
        self.reset_or_create_new_gesture()
    
    def reset_or_create_new_gesture(self):
        self.tempGestures = []
    
    def get_interpolated_gesture(self):
        if len(self.tempGestures) == 0:
            return None
        
        res = None
        for gesture in self.tempGestures:
            if not gesture.finalized:
                gesture.finalize(self.detaillevel)
            
            if not res:
                res = gesture
                continue
            
            res += gesture
            res.finalize(self.detaillevel)
        
        return res

class Builder(object):
    def __init__(self, database="", detaillevel=32):
        object.__init__(self)
        
        self.initialize_pygame()
        
        self.gestureManager = GestureManager(database, detaillevel)
        
        #initializes the required members
        self.reset_drawing_and_screen()
        
        self.leftPressed = False
        self.drawable = True
        
        # Dictionary that contains which fuction should be called,
        # if the events key is pressed. A handler gets the event and has to return True,,
        # if the program should run and False otherwise.
        self.handlerMapping = {
            pygame.K_ESCAPE: self.exit_program,
            pygame.K_q: self.exit_program,
            
            pygame.K_f: self.finalize_gesture,
            pygame.K_c: self.create_new_gesture,
            
            pygame.K_a: self.add_gesture,
            pygame.K_i: self.reset_drawing_and_screen,
            
            pygame.K_d: self.draw_temp_result,            
            pygame.K_p: self.print_result,
            pygame.K_e: self.print_all_gestures,
            
            pygame.K_w: self.watch_gesture_show,
            
            pygame.K_s: self.gestureManager.persist_gestures
        }
    
    def __del__(self):
        del self.gestureManager
    
    def exit_program(self, event=None):
        return False
    
    def watch_gesture_show(self, event=None):
        for gesture in self.gestureManager.gestures:
            self.draw_temp_result(interpolatedGesture=gesture)
            time.sleep(1)
            self.resultDrawn = False
        self.reset_drawing_and_screen()
        return True
    
    def finalize_gesture(self, event=None):
        self.gestureManager.finalize()
        return self.create_new_gesture(event)
    
    def create_new_gesture(self, event=None):
        self.gestureManager.reset_or_create_new_gesture()
        return self.reset_drawing_and_screen()
    
    def add_gesture(self, event=None):
        if not self.points:
            return
        
        gesture = Gesture(self.points)
        self.gestureManager.add_gesture(gesture)
        return self.reset_drawing_and_screen()
    
    def reset_drawing_and_screen(self, event=None):
        self.points = []        
        self.screen.fill((255,255,255))
        pygame.display.flip()
        self.drawable = True
        self.resultDrawn = False
        return True
    
    def draw_temp_result(self, event=None, interpolatedGesture=None):
        if self.resultDrawn:
            return True
        
        if not interpolatedGesture:
            interpolatedGesture = self.gestureManager.get_interpolated_gesture()
        
        if not interpolatedGesture:
            return True
        
        interpolatedGesture *= self.screen.get_height()/4
        interpolatedGesture += (self.screen.get_width()/2, self.screen.get_height()/2)
        
        self.reset_drawing_and_screen()
        self.points = interpolatedGesture.to_tuple_list()
        pygame.draw.circle(self.screen, (255, 128, 0), (int(self.points[0][0]),(int(self.points[0][1]))), 10)
        pygame.draw.lines(self.screen, (0, 200, 0), False, self.points, 2)
        pygame.display.flip()
        self.resultDrawn = True
        return True
    
    def print_result(self, event=None):
        gest = self.gestureManager.get_interpolated_gesture()
        if gest:
            print gest.to_tuple_list()
        
        return True
    
    def print_all_gestures(self, event=None):
        for x in xrange(10):
            print ""
        
        for gesture in self.gestureManager.gestures:
            print gesture.to_tuple_list()
            print ""
        
        return True
    
    def initialize_pygame(self):
        """Initializes the pygame sub system."""
        pygame.init()
        
        self.window = pygame.display.set_mode((1024,768), pygame.DOUBLEBUF)
        pygame.display.set_caption("warofthepolygons.com - Gesture Example")
        
        self.screen = pygame.display.get_surface()
   
    def handle_events(self):
        """Handles the input events.
        
        @return Returns True if the demo should run, False otherwise.
        """
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return False
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1 and self.drawable:
                    self.leftPressed = True
            elif event.type == pygame.MOUSEBUTTONUP:
                if event.button == 1:
                    self.leftPressed = False
                    if self.points:
                        self.drawable = False
            elif event.type == pygame.KEYDOWN:
                if event.key in self.handlerMapping:
                    return self.handlerMapping[event.key](event)
        
        return True
    
    def print_controls(self):
        print """F        - Finalize current gesture.
C        - Create new gesture. Current gesture should be finalized.

A        - Add current drawing to the gestures used for interpolation.
I        - Ignore current drawing for interpolation.

D        - Draw the interpolation result.
P        - Print the interpolation result as tuple list.
E        - Every interpolated gesture should be printed as list of tuples.

W        - Watch a slideshow of all finalized gestures.

S        - Save all gestures to the database.

ESC or Q - exit program

"""

    def run(self):
        print "CONTROLS\n========"
        self.print_controls()
        lastPointCount = -1
        
        while self.handle_events():
            if self.leftPressed:
                if self.resultDrawn:
                    self.reset_drawing_and_screen()
                    self.resultDrawn = False
                
                pos = pygame.mouse.get_pos()
                self.points.append(pos)
            
            if self.resultDrawn:
                continue
            
            if self.drawable:
                if len(self.points) > 1 and lastPointCount != len(self.points):
                    self.screen.fill((255, 255, 255))
                    pygame.draw.lines(self.screen, (0, 0, 255), False, self.points, 2)
                    lastPointCount = len(self.points)
                    pygame.display.flip()                

def __run__(database = ""):
    ## @todo use parameters from console here
    builder = Builder("..\data\gestures\gesture.db")
    builder.run()

def __run_performance__():
    import cProfile
    cProfile.run("__run__()", filename = "Builder.cprof")
    
    import pstats
    stats = pstats.Stats("Builder.cprof")
    stats.strip_dirs().sort_stats("time")
    stats.print_stats(10)

def _read_params():
    pass
    
if __name__ == "__main__":
    """Runs the demo and profiles its data."""
    __run__()