"""
############################################################################
# Copyright (c) 2009 T-002@gmx.de                                          #
#                                                                          #
# Permission is hereby granted, free of charge, to any person              #
# obtaining a copy of this software and associated documentation           #
# files (the "Software"), to deal in the Software without                  #
# restriction, including without limitation the rights to use,             #
# copy, modify, merge, publish, distribute, sublicense, and/or             #
# sell copies of the Software, and to permit persons to whom the           #
# Software is furnished to do so, subject to the following conditions:     #
#                                                                          #
# The above copyright notice and this permission notice                    #
# shall be included in all copies or substantial portions of the Software. #
#                                                                          #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,          #
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF       #
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   #
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY     #
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,     #
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE        #
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                   #
############################################################################
"""
__license__ = __doc__

import math

class Vector2D(object):
    """Represents a 2 dimensional vector."""
    
    def __init__(self, xOrTuple=(0,0), y=None):
        """Initialise the vector.
        
        @param xOrTuple Can be the x value or a tuple representing (x, y)
        @param y        Contains the y value, if xOrTuple is not a tuple.
        """
        object.__init__(self)
        
        if isinstance(xOrTuple, Vector2D):
            self.x = xOrTuple.x
            self.y = xOrTuple.y
        elif isinstance(xOrTuple, tuple):
            self.x = float(xOrTuple[0])
            self.y = float(xOrTuple[1])
        else:
            self.x = float(xOrTuple)
            self.y = float(y)
    
    def __str__(self):
        """Converts the vector to a string.
        
        @return returns a string containing "Vector2D(self.x, self.y)"
        """
        return "%s(%s, %s)" % (self.__class__.__name__, self.x, self.y)
    
    def __repr__(self):
        """Converts the Vector to its alternative representation as string.
        
        @return Returns a string containing "(self.x, self.y)"
        """
        return "(%s, %s)" % (self.x, self.y)

    def __mul__(self, item):
        """Multiplies the vector with the given item and returns the result.
        
        @param item Stuff used for multiplication. If this is an other Vector2D
                    the scalar product of the two vectors is returned, otherwise
                    the vector is scaled by item.
        
        @return Returns a number representing the scalar product, if item is an
                other Vector2D. Otherwise a new Vector2D instance is returned
                which represents the scaled vector.
        """
        if isinstance(item, Vector2D):
            return self.x * item.x + self.y * item.y
        else:
            return Vector2D(self.x * item, self.y * item)

    def div(self, divisor):
        """Scaled the vector down by the divisor.
        
        @param divisor Scaling divisor for the vector.
        
        @throw Can raise a ZeroDevisionError.
        """
        if divisor:
            self.x /= divisor
            self.y /= divisor
        else:
            self.x, self.y = 0.0, 0.0
        
        return self
    
    def __div__(self, divisor):
        """Scaled the factor down by the divisor.
        
        @param divisor Scaling divisor for teh factor.
        
        @return Returns a new Vector 2D representing the scaled vector.
        
        @throw Can raise a ZeroDevisionError.
        """
        return Vector2D(self.x / divisor, self.y / divisor)

    def sub(self, vec2D):
        """Used for vector substraction.
        
        @param vec2D Vector2D which should be substracted.
        """
        self.x -= vec2D[0]
        self.y -= vec2D[1]
        
        return self
    
    def __sub__(self, vec2D):
        """Used for vector substraction.
        
        @param vec2D Vector2D which should be substracted.
        
        @return Returns a new Vector2D representing the result.
        """
        return Vector2D(self.x - vec2D.x, self.y - vec2D.y)
    
    def add(self, vec2D):
        """Adds the given Vector2D instance to the vector.
        
        @param vec2D Vector2D instance which should be added.
        """
        self.x += vec2D[0]
        self.y += vec2D[1]
        
        return self
    
    def __add__(self, item):
        """Adds two Vector2D instances.
        
        @param item Vector2D or iterable instance which should be added.
        
        @return Returns a new Vector2D representing the result.
        """
        return Vector2D(self[0] + item[0], self[1] + item[1])
    
    def neg(self):
        """Multiplies the Vector with -1."""
        self.x = -self.x
        self.y = -self.y
        
        return self
    
    def __neg__(self):
        """Returns a negated version of the Vector2D.
        
        @return New Vector2D instance representing the negated vector.
        """
        return Vector2D(-self.y, -self.y)
    
    def __eq__(self, item):
        """Compares if item matches the Vector2D instance are equal to each other.
        
        @param item Can be another Vector2D or a tuple.
        
        @return Returns True, if the vectors are equal, False otherwise.
        """
        return self.x == item[0] and self.y == item[1]
    
    def __ne__(self, item):
        """Compares if to Vector2D instances are not equal to each other.
         
        @param item Can be another Vector2D or a tuple.
        
        @return Returns True, if the vectors are not equal, False otherwise.
        """
        return not self.__eq__(item)
    
    def __gt__(self, other):
        raise NotImplementedError("> is not defined for Vector2D")
    
    def __ge__(self, other):
        raise NotImplementedError(">= is not defined for Vector2D")
    
    def __lt__(self, other):
        raise NotImplementedError("< is not defined for Vector2D")
    
    def __le__(self, other):
        raise NotImplementedError("<= is not defined for Vector2D")
    
    def __abs__(self):
        """Returns the magnitude of the vector.
        
        @return Returns a float value which represents the vectors magnitude.
        """
        return math.hypot(self.x, self.y)

    def __getitem__(self, idx):
        if idx == 0:
            return self.x
        elif idx == 1:
            return self.y
        else:
            raise ArgumentError("Cannot evaluate iterator.")
    
    def __setitem__(self, idx, val):
        if idx == 0:
            self.x = val
        elif idx == 1:
            self.y = val
    
    def magnitude(self):
        """Returns the magnitude of the vector.
        
        @return Returns a float value which represents the vectors magnitude.
        """
        return abs(self)
    
    def normalized(self):
        """Returns a normalized version of the Vector2D.
        
        @return Returns a new Vector2D instance representing the normalized vector.
        """
        magnitude = self.magnitude()
        if magnitude == 0.0:
            return Vector2D(self.x, self.y)
        
        return Vector2D(self.x / magnitude, self.y / magnitude)
    
    def normalize(self):
        """Normalizes the vector."""
        magnitude = self.magnitude()
        
        if magnitude == 0.0:
            return
        
        self.x /= magnitude
        self.y /= magnitude
        
        return self
    
    def to_tuple(self):
        """Returns the tuple representation of the vector.
        
        @return Returns a tuple representing (x, y)
        """
        return (self.x, self.y)
